# Actually just many colors for plotting clusters
# The first then or so are nice, then it starts to get random

many = [
    "#15889C",
    "#ED496F",
    "#8E1382",
    "#FFB300",
    "#FF6F00",
    "#43A047",
    "#C62828",
    "#2d62a3",
    "#1F7530",
    "#573794",
    "#28AAE1",
    "#870f0f",
    "#9e9e9e",
    "#795548",
    "#CD6155",
    "#F7DC6F",
    "#7DCEA0",
    "#85C1E9",
    "#EB984E",
    "#ead9d5",
    "#03cd4A",
    "#CDDC39",
    "#e0594b",
    "#c76cde",
    "#24B177",
    "#8D6E63",
    "#486ff7",
    "#6300b5",
    "#88e200",
    "#012824",
    "#0d3290",
    "#a347fb",
    "#54fc7a",
    "#eb1388",
    "#b0978d",
    "#fe52cf",
    "#83f1f6",
    "#f1f847",
    "#2b1dfc",
    "#6c6f15",
    "#6ca05c",
    "#7788cd",
    "#f502f3",
    "#0dc290",
    "#fa0e03",
    "#3caa0a",
    "#befc8d",
    "#08f8eb",
    "#b1cd3f",
    "#d6a5fa",
    "#ce606c",
    "#ab1eba",
    "#6ecc9f",
    "#054ddc",
    "#486ff7",
    "#854f49",
    "#f22B21",
    "#3a0e43",
    "#225805",
    "#37d160",
    "#e4b974",
    "#a8bade",
    "#47edd1",
    "#f47a92",
    "#c76cde",
    "#9106eb",
    "#81aa20",
    "#d7fdfd",
    "#5deb2e",
    "#f82745",
    "#6435e0",
    "#027ffe",
    "#8e3101",
    "#16f648",
    "#1c15bc",
    "#8be46e",
    "#8d6fa0",
    "#e68fc6",
    "#058ca9",
    "#9e018a",
    "#bdfd0b",
    "#b22760",
    "#2bf49f",
    "#cb9348",
    "#9d8303",
    "#c251a1",
    "#46adaf",
    "#a3e3af",
    "#22bb34",
    "#6ea3fa",
    "#260374",
    "#1c3854",
    "#405d37",
    "#c21df3",
    "#fcea92",
    "#537f88",
    "#fd4c18",
    "#f2d71e",
    "#fd4c7a",
]


# Atlas color palette
atlas_colors = dict(
    {
        "neural_progenitor_cell": "#e9c46a",
        "telencephalic_npc": "#b77f92",
        "dorsal_npc": "#D19C2C",
        "dorsal_org": "#ffe082",
        "hippocampal_npc": "#F69B97",
        "ventral_npc": "#ce93d8",
        "ventral_org": "#9575cd",
        "dien_mesen_npc": "#80deea",
        "diencephalic_npc": "#29b6f6",
        "hypothalamic_npc": "#8EC7EC",
        "thalamic_npc": "#29D5F6",
        "mesencephalic_npc": "#26a69a",
        "retinal_npc": "#ffccbc",
        "rhombencephalic_npc": "#a5d6a7",
        "cerebellar_npc": "#95CB6E",
        "pons_npc": "#8BB599",
        "medullary_npc": "#BCD07E",
        "intermediate_progenitor": "#e76f51",
        "neuron": "#264653",
        "excitatory_neuron": "#A53259",
        "telencephalic_excitatory_neuron": "#b44a6e",
        "deeper_layer_cortical_excitatory_neuron": "#ec407a",
        "upper_layer_cortical_excitatory_neuron": "#8B1721",
        "hippocampal_excitatory_neuron": "#C35EA4",
        "non_telencephalic_excitatory_neuron": "#f9a825",
        "diencephalic_excitatory_neuron": "#0277bd",
        "thalamic_excitatory_neuron": "#2485BF",
        "hypothalamic_excitatory_neuron": "#134BBC",
        "mesencephalic_excitatory_neuron": "#00897b",
        "rhombencephalic_excitatory_neuron": "#558b2f",
        "cerebellar_excitatory_neuron": "#318C42",
        "pons_excitatory_neuron": "#4E703F",
        "medullary_excitatory_neuron": "#898B2F",
        "inhibitory_neuron": "#3A6890",
        "telencephalic_inhibitory_neuron": "#039be5",
        "ventral_inhibitory_neuron": "#76448a",
        "mge_inhibitory_neuron": "#8e44ad",
        "lge_inhibitory_neuron": "#5c6bc0",
        "cge_inhibitory_neuron": "#2B254B",
        "non_telencephalic_inhibitory_neuron": "#1B8466",
        "diencephalic_inhibitory_neuron": "#2196f3",
        "thalamic_inhibitory_neuron": "#5889A8",
        "hypothalamic_inhibitory_neuron": "#5A6EA8",
        "mesencephalic_inhibitory_neuron": "#00acc1",
        "rhombencephalic_inhibitory_neuron": "#4caf50",
        "cerebellar_inhibitory_neuron": "#699771",
        "pons_inhibitory_neuron": "#2F5D35",
        "medullary_inhibitory_neuron": "#C6B161",
        "choroid_plexus_epithelium": "#808b96",
        "astrocyte": "#DA2E7C",
        "oligodendrocyte_lineage": "#00bf99",
        "oligodendrocyte_precursor_cell": "#00FFCC",
        "mature_oligodendrocyte": "#00604d",
        "microglia": "#b2ff00",
        "vascular_endothelial_cell": "#F20000",
        "mesenchymal_cell": "#a1887f",
        "neural_crest": "#dce775",
        "pns_neurons": "#9e9d24",
    }
)
